#include <string.h>
#include <stdio.h>
#include "CommandQueue.h"
#include "log.h"

InternalCommandQueue::InternalCommandQueue() {
	currentIdx = 0;
	nextCommandIdx = 0;
	inData = false;
	suspended = false;
	inTemplate = false;
	inFunction = false;
}

CommandQueue::CommandQueue() : InternalCommandQueue() {
	mode = Append;
}

/**
 * ======== InternalCommandQueue::load ========
 *
 *  Purpose:
 *    Load all lines in the file as commands to be queued.
 *
 *  Input Parameters:
 *    fileName -- the file to load
 *
 *  Return Values:
 *    none
 */
void InternalCommandQueue::load( const char* filename ) {
	//log_message1( "InternalCommandQueue::load", filename );
	FILE* fp = fopen( filename, "r" );
	if ( fp != NULL ) {
		char line[1000];
		while ( fgets( &line[0], 1000, fp ) != NULL ) {
			if ( strlen( line ) > 0 ) {
				line[ strlen( line ) - 1 ] = 0;
			}
			queue( line );
		}
		fclose( fp );
	} else {
		log_message( "could not open file" );
	}
}


/**
 * ======== InternalCommandQueue::queue ========
 *
 *  Purpose:
 *    Parse a line of text, and queue up the appropriate command for
 *    execution.  This whole setup seems a bit odd, but is necessary
 *    because this thread cannot directly access VRML objects running
 *    in the other thread.
 * 
 *    This cross-thread access may be possible, but I don't know how
 *    to do it.  The obvious way does not work because of the COM
 *    implementation, which I do not understand.  Have to "marshall interface
 *    pointers" and use some special calls, no clear example online that
 *    I could find, ParallelGraphics people refer to MS documentation and 
 *    don't bother with examples at all.
 */

void InternalCommandQueue::queue( char* command ) {
	if ( debug ) {
		log_message1( "InternalCommandQueue::queue", command );
	}
	if ( inData ) {
		if ( strncmp( command, "enddata", 7 ) == 0 ) {
			addCommand( EndData, command );
			inData = 0;
		} else {		
			addCommand( Data, command );
		}
	} else if ( inTemplate ) {
		if ( strncmp( command, "endtemplate", 11 ) == 0 ) {
			//log_message( "Queueing endtemplate" );
			addCommand( EndTemplate, command );
			inTemplate = 0;
		} else {
			//log_message( "Queueing template data" );
			addCommand( TemplateData, command );
		}
	} else if ( inFunction ) {
		if ( strncmp( command, "endfunction", 11 ) == 0 ) {
			//log_message( "Queueing endtemplate" );
			addCommand( EndFunction, command );
			inFunction = 0;
		} else {
			//log_message( "Queueing template data" );
			addCommand( FunctionData, command );
		}

	} else if ( strncmp( command, "mswitch", 7 ) == 0 ) {
		addCommand( ModelSwitch, NULL );
	} else if ( strncmp( command, "rxcc", 4 ) == 0 ) {
		addCommand( RotateXcc, NULL );
	} else if ( strncmp( command, "rx", 2 ) == 0 ) {
		addCommand( RotateX, NULL );
	} else if ( strncmp( command, "rycc", 4 ) == 0 ) {
		addCommand( RotateYcc, NULL );
	} else if ( strncmp( command, "ry", 2 ) == 0 ) {
		addCommand( RotateY, NULL );
	} else if ( strncmp( command, "rzcc", 4 ) == 0 ) {
		addCommand( RotateZcc, NULL );
	} else if ( strncmp( command, "rz", 2 ) == 0 ) {
		addCommand( RotateZ, NULL );
	} else if ( strncmp( command, "fast", 4 ) == 0 ) {
		addCommand( FastMode, NULL );
	} else if ( strncmp( command, "slow", 4 ) == 0 ) {
		addCommand( SlowMode, NULL );
	} else if ( strncmp( command, "inlineattributes ", 17 ) == 0 ) {
		addCommand( TestDEF, command + 17 );
	} else if ( strncmp( command, "setinfo ", 8 ) == 0 ) {
		addCommand( SetInfo, command + 8 );
	} else if ( strncmp( command, "setproperty ", 12 ) == 0 ) {
		addCommand( SetVizProperty, command + 12 );
	} else if ( strncmp( command, "setvalue ", 9 ) == 0 ) {
		addCommand( SetVizValue, command + 9 );
	} else if ( strncmp( command, "clear", 5 ) == 0 ) {
		addCommand( Clear, NULL );
	} else if ( strncmp( command, "load ", 5 ) == 0 ) {
		addCommand( Load, command + 5 );
	} else if ( strncmp( command, "reset", 5 ) == 0 ) {
		addCommand( ReSetEngine, NULL );
	} else if ( strncmp( command, "show", 4 ) == 0 ) {
		addCommand( Show, NULL );
	} else if ( strncmp( command, "hide", 4 ) == 0 ) {
		addCommand( Hide, NULL );
	} else if ( strncmp( command, "left", 4 ) == 0 ) {
		addCommand( Left, NULL );
	} else if ( strncmp( command, "right", 5 ) == 0 ) {
		addCommand( Right, NULL );
	} else if ( strncmp( command, "inline ", 7 ) == 0 ) {
		addCommand( Inline, command + 7 );
	} else if ( strncmp( command, "in", 2 ) == 0 ) {
		addCommand( In, NULL );
	} else if ( strncmp( command, "out", 3 ) == 0 ) {
		addCommand( Out, NULL );
	} else if ( strncmp( command, "up", 2 ) == 0 ) {
		addCommand( Up, NULL );
	} else if ( strncmp( command, "down", 4 ) == 0 ) {
		addCommand( Down, NULL );
	} else if ( strncmp( command, "taller", 6 ) == 0 ) {
		addCommand( Taller, NULL );
	} else if ( strncmp( command, "shorter", 7 ) == 0 ) {
		addCommand( Shorter, NULL );
	} else if ( strncmp( command, "bigger", 6 ) == 0 ) {
		addCommand( Bigger, NULL );
	} else if ( strncmp( command, "smaller", 7 ) == 0 ) {
		addCommand( Smaller, NULL );
	} else if ( strncmp( command, "fatter", 6 ) == 0 ) {
		addCommand( Fatter, NULL );
	} else if ( strncmp( command, "thinner", 7 ) == 0 ) {
		addCommand( Thinner, NULL );
	} else if ( strncmp( command, "fadeout", 7 ) == 0 ) {
		addCommand( FadeOut, NULL );
	} else if ( strncmp( command, "fadein", 6 ) == 0 ) {
		addCommand( FadeIn, NULL );
	} else if ( strncmp( command, "setgroup", 8 ) == 0 ) {
		addCommand( SetGroup, NULL );
	} else if ( strncmp( command, "modelset ", 9 ) == 0 ) {
		addCommand( ModelSetCommand, command + 9 );
	} else if ( strncmp( command, "cleargroup", 10 ) == 0 ) {
		addCommand( ClearGroup, NULL );
	} else if ( strncmp( command, "cycle", 5 ) == 0 ) {
		addCommand( Cycle, NULL );
	} else if ( strncmp( command, "debugoff", 8 ) == 0 ) {
		debug = false;
	} else if ( strncmp( command, "debug", 5 ) == 0 ) {
		debug = true;
	} else if ( strncmp( command, "mark", 4 ) == 0 ) {
		if ( strncmp( command, "markx ", 6 ) == 0 ) {
			addCommand( MarkX, command + 6 );
		} else if ( strncmp( command, "marky ", 6 ) == 0 ) {
			addCommand( MarkY, command + 6 );
		} else if ( strncmp( command, "markz ", 6 ) == 0 ) {
			addCommand( MarkZ, command + 6 );
		}
	} else if ( strncmp( command, "setmax ", 7 ) == 0 ) {
		addCommand( SetMax, command + 7 );
	} else if ( strncmp( command, "label ", 6 ) == 0 ) {
		addCommand( Label, command + 6 );
	} else if ( strncmp( command, "infodimension ", 14 ) == 0 ) {
		addCommand( InfoDimension, command + 14 );
	} else if ( strncmp( command, "startdata ", 10 ) == 0 ) {
		addCommand( StartData, command + 10 );
		//log_message1("ADDING START DATA", command+10);
	} else if (( strncmp( command, "scale ", 6 ) == 0 ) && !inData ) {
		addCommand( Scale, command + 6 );
	} else if (strncmp( command, "setobject ", 10 ) == 0 ) {
		addCommand( SetObject, command + 10 );
	} else if ( strncmp( command, "starttemplate ", 14 ) == 0 ) {
		//log_message( "Queueing starttemplate" );
		inTemplate = 1;
		addCommand( StartTemplate, command + 14 );
	} else if ( strncmp( command, "startfunction ", 14 ) == 0 ) {
		inFunction = 1;
		addCommand( StartFunction, command + 14 );
	} else if ( strncmp( command, "setmarker", 9 ) == 0 ) {
		addCommand( SetMarker, NULL );
	} else if ( strncmp( command, "call ", 5 ) == 0 ) {
		addCommand( CallFunction, command + 5 );
	} else if ( strncmp( command, "spacex ", 7 ) == 0 ) {
		addCommand( SpaceX, command + 7 );
	} else if ( strncmp( command, "spacey ", 7 ) == 0 ) {
		addCommand( SpaceY, command + 7 );
	} else if ( strncmp( command, "spacez ", 7 ) == 0 ) {
		addCommand( SpaceZ, command + 7 );
	} else if ( strncmp( command, "link ", 5 ) == 0 ) {
		addCommand( Link, command + 5 );
	}
}

static char* process( char* s ) {
	int count = 0;
	while ( *s ) {
		if ((( *s == ':' ) || ( *s == ' ')) && ( count != 1 )) {
			*s = 0;
			s++;
			if ( *s == '_' ) {
				return( s + 1 );
			} else {
				return( s );
			}
		}
		s++;
		count++;
	}
	return( NULL );
}


void CommandQueue::addCommand( int cmd, const char* str ) {
	if ( mode == Insert ) {
		insertQueue.addCommand( cmd, str );
	} else {
		InternalCommandQueue::addCommand( cmd, str );
	}
}

/**
 * ======== InternalCommandQueue::addCommand ========
 *
 *  Purpose:
 *    Add a command to the command queue.  The type of command has already
 *    been parsed out and identified.
 *
 *  Input Parameters:
 *    cmd -- the type of command to add
 *    str -- parameters, if any to associate with the command.
 *
 *  Return Value:
 *    none
 */

void InternalCommandQueue::addCommand( int cmd, const char* str ) {
	if ( cmd == Clear ) { log_message( "INTERNAL QUEUE ADDING CLEAR" ); }
	if ( spaceAvailable() ) {
		command[ currentIdx ] = cmd;
		if (( cmd == StartData ) || ( cmd == ModelSetCommand )) {
			inData = 1;
			//log_message1( "CommandQueue:addCommand StartData of", str );
		} else if ( cmd == EndData ) {
			inData = 0;
		}
		if ( str == NULL ) {
			param1[ currentIdx ] = NULL;
			param2[ currentIdx ] = NULL;
		} else {
			param1[ currentIdx ] = newstrcpy( str );
			if ( cmd == ModelSetCommand ) {
				param2[ currentIdx ] = process( param1[ currentIdx ] );
			}

			// Label's get special handling, everything in label is param1
			if ( !inData && !inTemplate && ( cmd != Label ) && ( cmd != Inline )) {
				param2[ currentIdx ] = process( param1[ currentIdx ] );
			}
		}
		currentIdx = incValue( currentIdx );
	}
}

bool InternalCommandQueue::hasCommand() {
	if ( suspended ) {
		return( false );
	}
	if ( currentIdx != nextCommandIdx ) {
		return( true );
	} else {
		return( false );
	}
}

char* InternalCommandQueue::toStr( int command ) {
	switch ( command ) {
	case Clear:
		return( "Clear" );
	case Load:
		return( "Load" );
	case ReSetEngine:
		return( "ReSetEngine");
	case Show:
		return( "Show" );
	case Hide:
		return( "Hide" );
	case InfoDimension:
		return( "InfoDimension" );
	case SetInfo:
		return( "SetInfo" );
	case StartData:
		return( "StartData" );
	case EndData:
		return( "EndData" );
	case StartTemplate:
		return( "StartTemplate" );
	case EndTemplate:
		return( "EndTemplate" );
	case Data:
		return( "Data" );
	case Scale:
		return( "Scale" );
	case SetObject:
		return( "SetObject" );
	case Left:
		return( "Left" );
	case Right:
		return( "Right" );
	case Up:
		return( "Up" );
	case Down:
		return( "Down" );
	case In:
		return( "In" );
	case Out:
		return( "Out" );
	case Taller:
		return( "Taller" );
	case Shorter:
		return( "Bigger" );
	case Smaller:
		return( "Smaller" );
	case SetMax:
		return( "SetMax" );
	case SetVizValue:
		return( "SetVizValue" );
	case SetVizProperty:
		return( "SetVizProperty" );
	case Cycle:
		return( "Cycle" );
	case Fatter:
		return( "Fatter" );
	case Thinner:
		return( "Thinner" );
	case FadeOut:
		return( "FadeOut" );
	case FadeIn:
		return( "FadeIn" );
	case Inline:
		return( "Inline" );
	case Debug:
		return( "Debug" );
	case ModelSetCommand:
		return( "ModelSet" );
	case TestDEF:
		return( "TestDEF" );
	case FastMode:
		return( "FastMode" );
	case SlowMode:
		return( "SlowMode" );
	case SetGroup:
		return( "SetGroup" );
	case MarkX:
		return( "MarkX" );
	case MarkY:
		return( "MarkY" );
	case MarkZ:
		return( "MarkZ" );
	case CallFunction:
		return( "CallFunction" );
	default:
		log_messagei1( "unknown", command );
		return( "unknown" );
	}
}

Command* CommandQueue::getNextCommand() {
	if ( insertQueue.hasCommand() ) {
		return( insertQueue.getNextCommand() );
	} else {
		return( InternalCommandQueue::getNextCommand() );
	}
}

Command* InternalCommandQueue::getNextCommand() {
	if ( inData ) {
		//log_message( "no command" );
		return( NULL );
	}
	if ( hasCommand() ) {
		if ( debug ) 
			log_message1( "command:", toStr(command[ nextCommandIdx ]) );

		// If we have a SetVizValue command, skip all these until last one,
		// since they are absolute values and cancel each other out.
		//int testIdx = incValue( nextCommandIdx );
		//if ( command[ testIdx ] == SetVizValue ) {
		//	nextCommandIdx = testIdx;
		//}

		Command* result = new Command( command[ nextCommandIdx ], param1[ nextCommandIdx ], param2[ nextCommandIdx ] );
		nextCommandIdx = incValue( nextCommandIdx );
		return( result );
	} else {
		//log_message( "no command #2" );
		return( NULL );
	}

}

Command* CommandQueue::unconditionalGetNextCommand() {
	if ( insertQueue.hasCommand() ) {
		return( insertQueue.unconditionalGetNextCommand() );
	} else {
		return( InternalCommandQueue::unconditionalGetNextCommand() );
	}
}

Command* InternalCommandQueue::unconditionalGetNextCommand() {
	if ( hasCommand() ) {
		//log_message( "we have a command" );
		Command* result = new Command( command[ nextCommandIdx ], param1[ nextCommandIdx ], param2[ nextCommandIdx ] );
		nextCommandIdx = incValue( nextCommandIdx );
		return( result );
	} else {
		//log_message( "no command #2" );
		return( NULL );
	}

}

static int activeCommands = 0;
static int createCount = 0;
Command::Command( int c, char* p1, char* p2 ) {
	activeCommands++;
	createCount++;
	//if (( createCount % 100 ) == 0 ) {
	//	log_messagei1( "createCount: ",createCount);
	//	log_messagei1( "activeCount: ",activeCommands);
	//}
	command = c;
	param1 = p1;
	param2 = p2;
}

#include <malloc.h>
Command::~Command() {
	activeCommands--;
	if ( param1 != NULL ) {
		free( param1 );
	}
	/*if ( param2 != NULL ) {
		free( param2 );
	}*/
}

int InternalCommandQueue::incValue( int n ) {
	n++;
	if ( n == COMMAND_MAX ) {
		return( 0 );
	}
	return( n );
}

bool InternalCommandQueue::spaceAvailable() {
	if ( incValue( currentIdx ) != nextCommandIdx ) {
		return( true );
	} else {
		return( false );
	}
}
