#include <stdio.h>
#include "SceneTemplate.h"
#include "log.h"
#include "string.h"
#include "malloc.h"

/**
 * ======== SceneVariable::SceneVariable ========
 *
 *  Purpose:
 *    Container for most basic sort of renderer value that is used as a
 *    sort of environment-level variable for substituting into template.
 *    This can be used to hold things like installation location, or
 *    data values that get put into a template before it gets used.
 */
SceneVariable::SceneVariable( char* name, char* value ) {
	templateName = newstrcpy( name );
	if ( value == NULL ) {
		value = "";
	}
	templateChars = newstrcpy( value );
	//log_message1( "creating SceneVariable instance, named", name );
	//log_message1( "..associated characters", templateChars );
}


/**
 * ======== SceneVariable::setChars ========
 *
 *  Purpose:
 *    Set the value of an existing variable
 */

void SceneVariable::setChars( char* chars ) {
	free( templateChars );
	templateChars = newstrcpy( chars );
	//log_message1( "updating variable", templateName );
	//log_message1( "..associated characters", templateChars );
}


/**
 * ======== SceneVariable::~SceneVariable ========
 *
 *  Purpose:
 *    Free up the memory allocated by a SceneVariable
 */

SceneVariable::~SceneVariable()  {
	free( templateName );
	free( templateChars );
}


/**
 * ======== SceneTemplate::SceneTemplate ========
 *
 *  Purpose:
 *    Scene templates are VRML templates that get nodes substituted into them.
 *    This is to allow flexibility in how the shapes get treated, and to allow
 *    some level of programming to happen with the VRML.  This would allow the
 *    functionality of the renderer to changed based only on the external text.
 *
 *  Input Parameters:
 *    name -- the name of the scene template
 */

SceneTemplate::SceneTemplate( char* name ) : SceneVariable( name, "" ) {
	sceneTemplateLength = 0;
	mergedData = NULL;
}


/**
 * ======== SceneTemplate::~SceneTemplate ========
 *
 *  Purpose:
 *    Free up the memory allocated by a SceneTemplate
 */

SceneTemplate::~SceneTemplate() {
}


/**
 * ======== SceneTemplate::addLine ========
 *
 *  Purpose:
 *    Add a line to the scene template.  Expand the memory if necessary.
 *
 *  Input Parameters:
 *    line -- the line of text to add to the template.
 */

void SceneTemplate::addLine( char* line ) {
	unsigned int requiredLength = ((strlen( line ) + 2 + strlen( templateChars ) + 1)*100)/100;
	if ( requiredLength != sceneTemplateLength ) {
		sceneTemplateLength = requiredLength;
		char* newTemplate = (char*)malloc( sceneTemplateLength );
		strcpy( newTemplate, templateChars );
		strcat( newTemplate, line );
		free( templateChars );
		templateChars = newTemplate;
	} else {
		strcat( templateChars, line );
	}
	strcat( templateChars, "\n" );
}


/**
 * ======== SceneTemplate::mergeChars ========
 *
 *  Purpose:
 *    Allocate memory for a string, merge characters into it, and return.
 *
 *  Input Parameters:
 *    chars -- characters to merge into template
 *
 *  Return Value:
 *    allocated memory for the characters.  This memory remains allocated internally
 *    to this object.  It will get freed when it gets reused.
 */
#include <stdio.h>
char* SceneTemplate::mergeChars( char* chars ) {
	if ( mergedData != NULL ) {
		free( mergedData );
	}
	mergedData = (char*)malloc( strlen( chars ) + strlen( templateChars ) + 100 );
	sprintf( mergedData, templateChars, chars );
	return( mergedData );
}


/**
 * ======== SceneTemplate::mergeVariable ========
 *
 *  Purpose:
 *    Allocate memory for a string, find variable & merge value into right spot.
 *    Assumes expanded template already exists in "mergedData" from previous calls.
 *
 *  Input Parameters:
 *    variableName -- name of variable to search for in expandedTemplate
 *    value -- value of the variable
 *
 *  Return Value:
 *    allocated memory for the characters.  This memory remains allocated internally
 *    to this object.  It will get freed when it gets reused.
 */
#include <stdio.h>
char* SceneTemplate::mergeVariable( char* variableName, char* chars ) {
	char strToSearch[100];
	sprintf( strToSearch, "VAR_%s", variableName );
	if ( mergedData == NULL ) {
		mergedData = templateChars;
	}
	char* result = strstr( mergedData, strToSearch );
	if (( chars != NULL ) && ( result != NULL )) {
		//log_message1("MERGE VARIABLE", variableName );
		//log_message1("result is",result);
		//log_message1("chars are", chars);
		if ( mergedData == NULL ) {
			//log_message( "HEY, why is it NULL" );
			return( "" );
		}

		char* newresult = (char*)malloc( strlen( mergedData ) + strlen( chars ));
		*result = 0;
		strcpy( newresult, mergedData );
		//log_message1( "newresult A", newresult );
		strcat( newresult, chars );
		//log_message1( "newresult B", newresult );
		strcat( newresult, result + strlen( variableName ) + 4);
		//log_message1( "newresult C", newresult );
		free( mergedData );
		mergedData = newresult;
	}
	return( mergedData );
}



/**
 *  ======== SceneTemplateManager::SceneTemplateManager ========
 *
 *  Purpose:
 *    Manage a list of SceneTemplate objects.  Since I don't expect alot of
 *    these to exist, hard coded fixed limit of 200 (see header file).
 */

SceneTemplateManager::SceneTemplateManager() {
	templateIdx = 0;
}



/**
 * ======== SceneTemplate::scanVariables ========
 *
 *  Purpose:
 *    Start getting list of variables in the template, used with getNextVariable() below
 */

void SceneTemplate::scanVariables() {
	variablePointer = templateChars;
}


/**
 * ======== SceneTemplate::getNextVariable ========
 *
 *  Purpose:
 *    Get the next variable used in the template
 *
 *  Return Value:
 *    name of the next variable used in the template, NULL if 
 *    no more are left.
 */

char* SceneTemplate::getNextVariable() {
	static char result[100];
	while ( *variablePointer ) {
		if ( strncmp( variablePointer, "VAR_", 4 ) == 0 ) {
			variablePointer += 4;
			char* scanner = &result[0];
			while ( *variablePointer && (*variablePointer != ' ') && ( *variablePointer != '"')) {
				*scanner++ = *variablePointer++;
				*scanner = 0;
			}
			return( &result[0] );
		} else {
			variablePointer++;
		}
	}
	return( NULL );
}


/**
 * ======== SceneTemplateManager::getTemplateChars ========
 *
 *  Purpose:
 *    Look up the template characters by name
 *
 *  Input Parameters:
 *    name -- the name of the SceneTemplate to find
 *
 *  Return Values:
 *    The characters of the template, or NULL if the template is not found.
 */

char* SceneTemplateManager::getTemplateChars( char* name ) {
	SceneVariable* test = getVariable( name );
	//log_message1( "getTemplateChars for", name );
	if ( test != NULL ) {
		return( test->getChars() );
	} else {
		//log_message( "did not find it");
		return( NULL );
	}
}


/**
 * ========= SceneTemplateManager::getTemplate ========
 *
 *  Purpose:
 *    Look up a template by name
 *
 *  Input Parameters:
 *    name -- the name of the SceneTemplate to find
 *
 *  Return Value:
 *    A pointer to the SceneTemplate if found, otherwise NULL.
 */

SceneTemplate* SceneTemplateManager::getTemplate( char* name ) {
	for ( int i = 0; i < templateIdx; i++ ) {
		SceneTemplate* test = (SceneTemplate*)templateList[i];
		if ( strcmp( name, test->getName() ) == 0 ) {
			return( (SceneTemplate*)test );
		}
	}
	return( NULL );
}


/**
 * ========= SceneTemplateManager::getVariable ========
 *
 *  Purpose:
 *    Look up a variable by name
 *
 *  Input Parameters:
 *    name -- the name of the SceneVariable to find
 *
 *  Return Value:
 *    A pointer to the SceneTemplate if found, otherwise NULL.
 */

SceneVariable* SceneTemplateManager::getVariable( char* name ) {
	//log_message1( "SceneTemplateManager::getVariable", name );
	for ( int i = 0; i < templateIdx; i++ ) {
		SceneVariable* test = templateList[i];
		//log_message1( "compare to", test->getName() );
		if ( strcmp( name, test->getName() ) == 0 ) {
			//log_message( "FOUND IT" );
			//log_message( test->getChars() );
			return( test );
		}
	}
	//log_message( "DID NOT FIND IT" );
	return( NULL );
}


/**
 * ======== SceneTemplateManager::addTemplate ========
 *
 *  Purpose:
 *    Create a new SceneTemplate and add it to the list.
 *
 *  Input Parameters:
 *    name -- the name of the template.  All scene templates are accessed by name
 */

SceneTemplate* SceneTemplateManager::addTemplate( char* name ) {
	//log_message1( "SceneTemplateManager::addTemplate", name );
	SceneTemplate* s = new SceneTemplate( name );
	templateList[ templateIdx++ ] = s;
	return( s );
}


/**
 * ======== SceneTemplateManager::addVariable ========
 *
 *  Purpose:
 *    Create a new SceneVariable and add it to the list.
 *
 *  Input Parameters:
 *    name -- the name of the Variable.  All scene variables are accessed by name
 *    value -- the value associated with the variable
 *
 *  Return Values:
 *     the SceneVariable containing the name-value association.  The caller manages
 *     the "value" memory, the SceneVariable makes its own copy.
 */

SceneVariable* SceneTemplateManager::addVariable( char* name, char* value ) {
	//log_message1( "SceneTemplateManager::addVariable", name );
	//log_message1( "value is", value );
	SceneVariable* test = getVariable( name );
	if ( test != NULL ) {
		test->setChars( value );
		return( test );
	} else {
		SceneVariable* s = new SceneVariable( name, value );
		templateList[ templateIdx++ ] = s;
		return( s );
	}
}


/**
 * ======== SceneTemplateManager::addLine ========
 *
 *  Purpose:
 *    Add a line of text to the most recently added SceneTemplate
 *
 *  Input Parameters:
 *    line -- the line of text to add
 */

void SceneTemplateManager::addLine( char* line ) {
	//log_message1( "SceneTemplateManager::addLine", line );
	SceneTemplate* latest = (SceneTemplate*)templateList[ templateIdx - 1 ];
	//log_message1( "latest template name is: ", latest->getName() );
	latest->addLine( line );
}
